@@ -1,3 +1,4 @@
+
 /*
  *  GTetrinet
  *  Copyright (C) 1999, 2000, 2001, 2002, 2003  Ka-shu Wong (kswong@zip.com.au)
@@ -68,6 +69,11 @@ int spectatorcount = 0;
 char specialblocks[256];
 int specialblocknum = 0;
 
+/*
+ * this will have the number of /list commands sended and waiting for answer
+ */
+gint list_issued;
+
 FIELD fields[7];
 
 int moderator; /* are we the moderator ? TRUE : FALSE */
@@ -133,6 +139,8 @@ struct sb sbinfo[] = {
     {0, 0, 0}
 };
 
+static guint up_chan_list_source;
+
 static void tetrinet_updatelevels (void);
 static void tetrinet_setspeciallabel (int sb);
 static void tetrinet_dospecial (int from, int to, int type);
@@ -175,11 +183,12 @@ void tetrinet_inmessage (enum inmsg_type msgtype, char *data)
     /* process the message */
     switch (msgtype) {
     case IN_CONNECT:
-        /* do nothing - setup is done with the playernum message */
+        list_issued = 0;
+        up_chan_list_source = g_timeout_add (30000, (GSourceFunc) partyline_update_channel_list, NULL);
         break;
     case IN_DISCONNECT:
         if (!connected) {
-            data = "Server disconnected";
+            data = _("Server disconnected");
             goto connecterror;
         }
         if (ingame) tetrinet_endgame ();
@@ -195,11 +204,14 @@ void tetrinet_inmessage (enum inmsg_type msgtype, char *data)
                 playernames[i][0] = teamnames[i][0] = 0;
             fieldslabelupdate ();
         }
+        g_source_remove (up_chan_list_source);
         winlist_clear ();
         fields_attdefclear ();
         fields_gmsgclear ();
         partyline_fmt (_("%c%c*** Disconnected from server"),
                        TETRI_TB_C_DARK_GREEN, TETRI_TB_BOLD);
+        partyline_clear_list_channel ();
+        partyline_joining_channel (NULL);
         break;
     case IN_CONNECTERROR:
     connecterror:
@@ -292,6 +304,8 @@ void tetrinet_inmessage (enum inmsg_type msgtype, char *data)
             partylineupdate_join (playernames[pnum]);
             /* check moderator status */
             checkmoderatorstatus ();
+            /* update channel list */
+            partyline_update_channel_list ();
             /* send out our field */
             /* if (!spectating) tetrinet_resendfield (); */
         }
@@ -391,13 +405,73 @@ void tetrinet_inmessage (enum inmsg_type msgtype, char *data)
                     tetrix = TRUE;
                 }
                 else if (tetrix) {
+                    gchar *line = nocolor (token);
+                      
+                    if (list_issued > 0)
+                    {
+                      if (*line == '(')
+                      {
+                        partyline_add_channel (line);
+                        break;
+                      }
+                      
+                      if (!strncmp ("List", line, 4))
+                      {
+                        partyline_more_channel_lines ();
+                        break;
+                      }
+                      
+                      if (!strncmp ("TetriNET", line, 8))
+                        break;
+                      
+                      if (!strncmp ("You", line, 3))
+                      {
+                        /* we will use the error message as list stopper */
+                        list_issued--;
+                        if (list_issued <= 0)
+                          stop_list();
+                        break;
+                      }
+                      
+                      //if (line != NULL) g_free (line);
+                    }
+                    /* detect whenever we have joined a channel */
+                    if (!strncmp ("has joined", &line[strlen (playernames[playernum])+1], 10))
+                    {
+                      partyline_joining_channel (&line[strlen (playernames[playernum])+20]);
+                    }
+                    else if (!strncmp ("Joined existing Channel", line, 23))
+                    {
+                      partyline_joining_channel (&line[26]);
+                    }
+                    else if (!strncmp ("Created new Channel", line, 19))
+                    {
+                      partyline_joining_channel (&line[22]);
+                    }
+                    
                     g_snprintf (buf, sizeof(buf), "*** %s", token);
                     partyline_text (buf);
                     break;
                 }
-                else plinemsg ("Server", token);
+                else
+                  plinemsg ("Server", token);
+            }
+            else
+            {
+              if (pnum == playernum)
+              {
+                gchar *line = nocolor (token);
+                if (!strncmp (line, "(msg) --- MARK ---", 18))
+                {
+                  list_issued--;
+                  if (list_issued <= 0)
+                    stop_list();
+                }
+                //g_free (line);
+              }
+              else
+                plinemsg (playernames[pnum], token);
             }
-            else plinemsg (playernames[pnum], token);
         }
         break;
     case IN_PLINEACT:
@@ -761,6 +835,10 @@ void tetrinet_playerline (const char *text)
             client_outmessage (OUT_PLINE, buf);
             return;
         }
+        /* send the message without showing it in the partyline */
+        g_snprintf (buf, sizeof(buf), "%d %s", playernum, text);
+        client_outmessage (OUT_PLINE, buf);
+        return;
     }
     g_snprintf (buf, sizeof(buf), "%d %s", playernum, text);
     client_outmessage (OUT_PLINE, buf);
@@ -784,7 +862,10 @@ void tetrinet_changeteam (const char *newteam)
     }
 }
 
-void tetrinet_sendfield (int reset)
+#if 0
+#define TETRINET_CHECK_COMPAT_SENDFIELD(x) /* do nothing */
+#else
+static void tetrinet_old_sendfield (int reset, const char *new_buf)
 {
     int x, y, i, d = FALSE;
     char buf[1024], buf2[1024], *p;
@@ -821,10 +902,84 @@ void tetrinet_sendfield (int reset)
         *p = 0;
         g_snprintf (buf, sizeof(buf), "%d %s", playernum, buf2);
     }
+    if (strcmp(new_buf, buf))
+      g_warning("sendfield ... non compat. data\n"
+                "\tBEG: sendfield data\n"
+                "\t old=%s\n"
+                "\t new=%s\n"
+                "\tEND: sendfield data\n",
+                buf, new_buf);
+#if 0
     /* send it */
     client_outmessage (OUT_F, buf);
     /* update the one in our memory */
     copyfield (sentfield, fields[playernum]);
+#endif
+}
+#define TETRINET_CHECK_COMPAT_SENDFIELD(x) tetrinet_old_sendfield (reset, (x))
+#endif
+
+void tetrinet_sendfield (int reset)
+{
+  int x, y, i, d = 0; /* d is the number of differences */
+  char buf[1024], *p;
+
+  char diff_buf[15][(FIELDWIDTH + 1)* FIELDHEIGHT * 2] = {0};
+
+  int row_count[15] = {1,1,1,1,1, 1,1,1,1,1, 1,1,1,1,1};
+  
+  g_snprintf (buf, sizeof(buf), "%d ", playernum);
+
+  if(!reset) {
+    /* Find out which blocks changed, how, and how many */
+    for (y = 0; y < FIELDHEIGHT; y ++) {
+      for (x = 0; x < FIELDWIDTH; x ++) {
+
+	const int block = fields[playernum][y][x];
+
+        if ((block < 0) || (block >= 15))
+        {
+          g_warning("sendfield shouldn't reach here, block=%d\n", block);
+          continue;
+        }
+        
+	if (block != sentfield[y][x]) {
+	  diff_buf[block][row_count[block]++] = x + '3';
+	  diff_buf[block][row_count[block]++] = y + '3';
+	  d += 2;
+	}
+      }
+    }
+    if (d == 0) return; /* no differences */
+
+    for (i = 0; i < 15; ++i)
+      if (row_count[i] > 1)
+        ++d; /* add an extra value for the '!'+i at the start */
+  }
+  
+  if (reset || d >= (FIELDHEIGHT*FIELDWIDTH)) {
+    /* sending entire field is more efficient */
+    p = buf + 2;
+    for (y = 0; y < FIELDHEIGHT; y ++)
+      for (x = 0; x < FIELDWIDTH; x ++)
+	*p++ = blocks[(int)fields[playernum][y][x]];
+    *p = 0;
+  }
+  else { /* so now we need to create the buffer strings */
+    for(i = 0; i < 15; ++i) {
+      if(row_count[i] > 1) {
+	diff_buf[i][0] = '!' + i;
+	GTET_O_STRCAT (buf, diff_buf[i]);
+      }
+    }
+  }
+
+  TETRINET_CHECK_COMPAT_SENDFIELD(buf);
+  
+  /* send it */
+  client_outmessage (OUT_F, buf);
+  /* update the one in our memory */
+  copyfield (sentfield, fields[playernum]);
 }
 
 void tetrinet_updatefield (FIELD field)
@@ -1520,47 +1675,6 @@ int tetrinet_key (int keyval)
     }
 notfieldkey:
     if (!ingame) return FALSE;
-    /* gmsg keys */
-/*    if (gmsgstate == 1) {
-        fields_gmsginputactivate (TRUE);
-        switch (keyval) {
-        case GDK_Return:
-            {
-                char buf[256];
-                const char *s;
-                gchar *locale_s;
-                s = fields_gmsginputtext ();
-                if (strlen(s) > 0) {
-                    if (strncmp("/me ", s, 4) == 0) {
-                        g_snprintf (buf, sizeof(buf), "* %s %s", nick, s+4);
-                        locale_s = g_locale_from_utf8 (buf, -1, NULL, NULL, NULL);
-                        client_outmessage (OUT_GMSG, locale_s);
-                    }
-                    else {
-                        g_snprintf (buf, sizeof(buf), "<%s> %s", nick, s);
-                        locale_s = g_locale_from_utf8 (buf, -1, NULL, NULL, NULL);
-                        client_outmessage (OUT_GMSG, locale_s);
-                    }
-                    g_free (locale_s);
-                }
-                fields_gmsginput (FALSE);
-                fields_gmsginputclear ();
-                gmsgstate = 0;
-            }
-            break;
-        case GDK_Escape:
-            gmsgstate = 2;
-            fields_gmsginput (FALSE);
-            break;
-        case GDK_BackSpace:
-            fields_gmsginputback ();
-            break;
-        default:
-            fields_gmsginputadd (str);
-            break;
-        }
-        return TRUE;
-    }*/
     if (gdk_keyval_to_lower (keyval) == keys[K_GAMEMSG]) {
         fields_gmsginput (TRUE);
         gmsgstate = 1;
